--- c:/checkout/emacs-cvs/emacs/src/w32term.c	Tue Feb 20 02:22:42 2007
+++ w32term.c	Wed Feb 21 00:10:48 2007
@@ -3468,11 +3468,70 @@
 /* Set the thumb size and position of scroll bar BAR.  We are currently
    displaying PORTION out of a whole WHOLE, and our position POSITION.  */
 
+#define SCROLL_FIX
+
 static void
 w32_set_scroll_bar_thumb (bar, portion, position, whole)
      struct scroll_bar *bar;
      int portion, position, whole;
 {
+#ifdef SCROLL_FIX
+    Window w;
+    BOOL draggingp = !NILP (bar->dragging) ? TRUE : FALSE;
+    SCROLLINFO si;
+    int sb_page, sb_pos, range, window, lines, n, mem;
+
+    static int was_draggingp;
+    if (draggingp && was_draggingp)
+        return;
+    was_draggingp = draggingp;
+
+    // scan some text
+    mem = whole;
+    if (mem > 20000)
+        mem = 20000;
+
+    for (lines = 0, n = 1; n <= mem; ++n)
+        if (FETCH_BYTE(n) == '\n')
+            ++lines;
+
+    window = WINDOW_TOTAL_LINES (XWINDOW (bar->window));
+    range = VERTICAL_SCROLL_BAR_INSIDE_HEIGHT (f, XINT(bar->height));
+
+    if (mem == whole) {
+        sb_page = range * window / (lines + window);
+    } else {
+        window = window * mem / (lines + 1); // in bytes, avg.
+        sb_page = range * window / (whole + window);
+    }
+
+    // bigger is nicer for them with a cheap mouse
+    n = VERTICAL_SCROLL_BAR_INSIDE_WIDTH (f, XINT(bar->width));
+    if (sb_page < n) {
+        sb_page = n;
+        // but not too big ...
+        n = range/4;
+        if (sb_page > n)
+            sb_page = n;
+    }
+
+    if (0 == whole)
+        sb_pos = 0;
+    else
+        sb_pos = ((double) position * (range - sb_page) + whole - 1) / whole;
+
+    si.cbSize = sizeof (si);
+    si.fMask = SIF_PAGE | SIF_POS | SIF_RANGE | SIF_DISABLENOSCROLL;
+    si.nMin = 0;
+    si.nMax = range - 1;
+    si.nPage = sb_page;
+    si.nPos = sb_pos;
+
+    w = SCROLL_BAR_W32_WINDOW (bar);
+    SetScrollInfo (w, SB_CTL, &si, TRUE);
+    return;
+
+#else
   Window w = SCROLL_BAR_W32_WINDOW (bar);
   /* We use the whole scroll-bar height in the calculations below, to
      avoid strange effects like scrolling backwards when just clicking
@@ -3538,6 +3597,7 @@
   SetScrollInfo (w, SB_CTL, &si, TRUE);
 
   UNBLOCK_INPUT;
+#endif
 }
 
 
@@ -3946,9 +4006,15 @@
     si.cbSize = sizeof (si);
     si.fMask = SIF_POS;
 
+#ifdef SCROLL_FIX
+    si.fMask = SIF_POS|SIF_RANGE|SIF_PAGE; //gr
+#endif
     GetScrollInfo ((HWND) msg->msg.lParam, SB_CTL, &si);
     y = si.nPos;
 
+#ifdef SCROLL_FIX
+    top_range = si.nMax + 1 - si.nPage; //gr
+#endif
     bar->dragging = Qnil;
 
 
@@ -3983,6 +4049,7 @@
 	bar->dragging = Qt;
 	emacs_event->part = scroll_bar_handle;
 
+#ifndef SCROLL_FIX
 	/* "Silently" update current position.  */
 	{
 	  SCROLLINFO si;
@@ -3996,11 +4063,13 @@
 
 	  SetScrollInfo (SCROLL_BAR_W32_WINDOW (bar), SB_CTL, &si, FALSE);
 	}
+#endif
 	break;
       case SB_ENDSCROLL:
 	/* If this is the end of a drag sequence, then reset the scroll
 	   handle size to normal and do a final redraw.  Otherwise do
 	   nothing.  */
+#ifndef SCROLL_FIX
 	if (dragging)
 	  {
 	    SCROLLINFO si;
@@ -4013,6 +4082,7 @@
 	    si.nPos = last_scroll_bar_drag_pos;
 	    SetScrollInfo (SCROLL_BAR_W32_WINDOW (bar), SB_CTL, &si, TRUE);
 	  }
+#endif
 	/* fall through */
       default:
 	emacs_event->kind = NO_EVENT;
